package se.chalmers.paragon;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;

/**
 * <p>
 * Implements a Policy.
 * </p>
 * 
 * @author Paragon Developers (http://www.cse.chalmers.se/paragon)
 * 
 */
public class Policy {

	/**
	 * The name of this policy (for messages)
	 */
	private String policyName;

	private Lock myFlow;

	/**
	 * Keeping track of all registered policies
	 */
	private static HashSet<String> existingNames;

	/**
	 * Constructs a new policy
	 * 
	 * @param policyName
	 *            The name of this policy
	 */
	protected Policy(String policyName) {
		if (existingNames == null)
			existingNames = new HashSet<String>();
		if (existingNames.contains(policyName))
			throw new UniqueNameError("Policy name " + policyName
					+ " is not unique.");
		this.policyName = policyName;
		this.myFlow = Lock.freshFlowLock(policyName);
	}


	/**
	 * Returns a new policy
	 * 
	 * @param policyName
	 *            Name of the policy
	 * @return
	 */
	public static Policy newPolicy(String policyName, PClause... clauses) {
		Policy p = new Policy(policyName);
		for (PClause c : clauses) {
			p.addPClause(c);
		}
		return p;
	}

	/**
	 * Creates a new policy with a fresh, unused name
	 * 
	 * @param base
	 * @return
	 */
	private static Policy freshPolicy(String base) {
		int i = 0;
		String name = base;
		while (existingNames.contains(name)) {
			name = base + i;
			i += 1;
		}
		return new Policy(name);
	}

	/**
	 * The name of this policy
	 * 
	 * @return
	 */
	public String getName() {
		return this.policyName;
	}

    private void addClause(ActorList head, Atom... body) {
		this.myFlow.addClause(head,body);
	}

    public static PClause newPClause(Actor head, Atom... body) {
    	return new PClause(head, body);
    }

    public void addPClause(PClause pc) {
    	this.myFlow.addClause(ActorList.newActorList(pc.getHead()), pc.getBody());
    }
    
	/**
	 * 
	 * Acts as the 'is at least as restrictive as' comparison between policies
	 * 
	 * @param other
	 *            Another policy
	 * @return True if every flow allowed by this policy is also allowed by the
	 *         other policy.
	 */
	public boolean implies(Policy other) {

		for (Clause clause : this.getRelatedClauses()) {

			// Create substitution
			HashMap<ActorVariable, ConcreteActor> varmap = new HashMap<ActorVariable, ConcreteActor>();
			for (Atom l : clause.getBody()) {
				varmap = Util.skolemVars(l, varmap);
			}

			// Substitute in head and body
			Atom head = Util.substitute(clause.getHead(), varmap);
			Atom[] body = new Atom[clause.getBody().length];
			for (int i = 0; i < body.length; i++)
				body[i] = Util.substitute(clause.getBody()[i], varmap);

			// open all locks in body, keep track of the locks already open (to
			// close the correct ones again) When the lock is 'this', i.e. a
			// recursive lock, open the one from policy instead.
			for (Atom lock : body) {
				lock.getLock()
						.openFreshFact(lock.getArgs().getActorsConcrete());
			}

			// Try and derive head for the other policy
			boolean canDerive = other.canDerive(head.getArgs()
					.getActorsConcrete());

			// Close locks again
			for (Atom lock : body)
				lock.getLock().clearFreshFacts();
			// If failed, return false
			if (!canDerive)
				return false;
		}
		// Could make same derivations for all clauses
		return true;
	}
	
	/**
	 * Checks whether the lock is open for the provided arguments. When the
	 * explicit state does not have this lock registered as open, relevant lock
	 * clauses are applied the implicit state until either
	 * <ul>
	 * <li>The lock is derived to be open for the arguments
	 * <li>The entire implicit state is derived and the lock is not open for
	 * this instance
	 * </ul>
	 * 
	 * @param actors
	 *            List of concrete actors to check the state of this lock for
	 * @return Returns true if this lock is open, either explicit or via clauses
	 */
	public boolean canDerive(ConcreteActor... actors) {
		return this.myFlow.isOpen(actors);
	}

	/**
	 * Creates a new lock that copies both the clauses from this clause and the
	 * provided argument with the new lock as head. In this way it is
	 * effectively the meet for policies, i.e. the returned lock is open for
	 * some actors if this lock or the lock provided as argument is open for
	 * these actors.
	 * 
	 * @param policy
	 *            The lock (policy) to meet with
	 * @return The result of applying the meet operation on locks.
	 */
	public Policy meet(Policy policy) {

		// Fresh name:
		Policy result = Policy.freshPolicy(
				"meet" + this.getName() + policy.getName());

		// Copy clauses:
		for (Clause r : this.getRelatedClauses())
			result.addClause(r.getHead().getArgs(), r.getBody());
		for (Clause r : policy.getRelatedClauses())
			result.addClause(r.getHead().getArgs(), r.getBody());
		return result;
	}

	/**
	 * Creates a new lock with clauses such that the new lock is open for some
	 * actors if it is open for <i>both</i> this lock and the lock provided as
	 * argument. This is effectively the join for policies.
	 * 
	 * @param policy
	 *            The lock (policy) to join with
	 * @return The result of applying the join operation on locks.
	 */
	public Policy join(Policy policy) {

		// Fresh name:
		Policy result = Policy.freshPolicy(
				"join" + this.getName() + policy.getName());

		for (Clause thisclause : this.getRelatedClauses()) {
			Atom thishead = thisclause.getHead();
			for (Clause otherclause : policy.getRelatedClauses()) {

				Atom otherhead = otherclause.getHead();

				// Get a mapping so that this head agrees with other head on the
				// concrete actors
				HashMap<ActorVariable, ConcreteActor> varmap = Util
						.sameConcreteActors(thishead, otherhead);

				// If such a mapping cannot be found, try the next combination
				if (varmap == null)
					continue;

				// Perform the substitution on this head, now all concrete
				// actors from the other head are at the same spot in this head
				thishead = Util.substitute(thishead, varmap);
				Atom[] thisbody = Util
						.substitutes(thisclause.getBody(), varmap);

				// The other way around, find a mapping from such that the other
				// head agrees with the concrete variables. This resulting
				// mapping can never be null. (Right?)
				varmap = Util.sameConcreteActors(otherhead, thishead);

				// Apply this mapping on the other clause - now both heads have
				// exactly the same concrete actors at the same place.
				otherhead = Util.substitute(otherhead, varmap);
				Atom[] otherbody = Util.substitutes(otherclause.getBody(),
						varmap);

				// Now we need to make sure they give the same naming to
				// variables. Get the highest variable ID so that we can create
				// fresh ones.
				int highestVar = Util.getHighestVar(thisbody);
				highestVar = Math.max(highestVar,
						Util.getHighestVar(otherclause.getBody()));

				// Map this heads variable to fresh one - generated sequential
				// so that they align with the ones for the other head
				HashMap<ActorVariable, ActorVariable> varvarmap = Util
						.createVarVarMap(thishead, highestVar + 1);
				thishead = Util.substituteVar(thishead, varvarmap);
				thisbody = Util.substitutesVar(thisbody, varvarmap);

				// Map the other heads variable to fresh one - generated
				// sequential so that they align with the ones for this head
				varvarmap = Util.createVarVarMap(otherhead, highestVar + 1);
				otherhead = Util.substituteVar(otherhead, varvarmap);
				otherbody = Util.substitutesVar(otherbody, varvarmap);

				// Otherhead should now be equal to thishead, except for the
				// name of the lock. The bodies are updated accordingly.

				// Generate the new body for this clause
				Atom[] newbody = new Atom[otherbody.length + thisbody.length];
				System.arraycopy(thisbody, 0, newbody, 0, thisbody.length);
				System.arraycopy(otherbody, 0, newbody, thisbody.length,
						otherbody.length);
				result.addClause(thishead.getArgs(), newbody);
			}
		}

		return result;
	}


	public String toString() {
		String res = getName();
		res += " {\n";
		for (Clause r : this.getRelatedClauses())
			res += "   " + r.toString() + "\n";
		res += "  }";
		return res;
	}

	private LinkedList<Clause> getRelatedClauses() {
		return this.myFlow.getRelatedClauses();
	}

}
